import nbformat as nbf
from typing import Any
import bencher as bch


from bencher.example.meta.example_meta import BenchableObject


class BenchMetaGen(bch.ParametrizedSweep):
    """This class uses bencher to display the multidimensional types bencher can represent"""

    float_vars = bch.IntSweep(
        default=0, bounds=(0, 4), doc="The number of floating point variables that are swept"
    )
    categorical_vars = bch.IntSweep(
        default=0, bounds=(0, 3), doc="The number of categorical variables that are swept"
    )
    sample_with_repeats = bch.IntSweep(default=1, bounds=(1, 100))

    sample_over_time = bch.BoolSweep(default=False)

    level = bch.IntSweep(default=2, units="level", bounds=(2, 5))

    run_bench = False

    plots = bch.ResultReference(units="int")

    def __call__(self, **kwargs: Any) -> Any:
        self.update_params_from_kwargs(**kwargs)

        run_cfg = bch.BenchRunCfg()
        run_cfg.level = self.level
        run_cfg.repeats = self.sample_with_repeats
        run_cfg.over_time = self.sample_over_time
        run_cfg.plot_size = 500

        # bench = bch.Bench("benchable", BenchableObject(), run_cfg=run_cfg)

        inputs_vars_float = [
            "float1",
            "float2",
            bch.p("float3", max_level=3),
            "sigma",
        ]

        inputs_vars_cat = [
            "noisy",
            "noise_distribution",
            "negate_output",
        ]

        input_vars = inputs_vars_float[: self.float_vars] + inputs_vars_cat[: self.categorical_vars]

        if self.run_bench:
            bench = BenchableObject().to_bench(run_cfg)
            res = bench.plot_sweep(
                "test",
                input_vars=input_vars,
                result_vars=["distance", "sample_noise"],
                plot_callbacks=False,
            )
            self.plots = bch.ResultReference()
            self.plots.obj = res.to_auto()

        title = f"{self.float_vars}_float_{self.categorical_vars}_cat"

        nb = nbf.v4.new_notebook()
        text = f"""# {title}"""

        code_gen = f"""
%%capture
import bencher as bch
from bencher.example.meta.example_meta import BenchableObject

run_cfg = bch.BenchRunCfg()
run_cfg.repeats = {self.sample_with_repeats}
run_cfg.level = 4 
bench = BenchableObject().to_bench(run_cfg)
res=bench.plot_sweep(input_vars={input_vars},
                    result_vars=["distance","sample_noise"])
"""
        code_results = """
from bokeh.io import output_notebook
output_notebook()
res.to_auto_plots()
"""

        nb["cells"] = [
            nbf.v4.new_markdown_cell(text),
            nbf.v4.new_code_cell(code_gen),
            nbf.v4.new_code_cell(code_results),
        ]
        from pathlib import Path

        fname = Path(f"docs/reference/Meta/ex_{title}.ipynb")
        fname.parent.mkdir(parents=True, exist_ok=True)
        fname.write_text(nbf.writes(nb), encoding="utf-8")

        return super().__call__()


def example_meta(run_cfg: bch.BenchRunCfg = None, report: bch.BenchReport = None) -> bch.Bench:
    bench = BenchMetaGen().to_bench(run_cfg, report)

    bench.plot_sweep(
        title="Meta Bench",
        description="""## All Combinations of Variable Sweeps and Resulting Plots
This uses bencher to display all the combinations of plots bencher is able to produce""",
        input_vars=[
            bch.p("float_vars", [0, 1]),
            "categorical_vars",
            bch.p("sample_with_repeats", [1, 20]),
        ],
        const_vars=[
            # BenchMeta.param.float_vars.with_const(1),
            # BenchMeta.param.sample_with_repeats.with_const(2),
            # BenchMeta.param.categorical_vars.with_const(2),
            # BenchMeta.param.sample_over_time.with_const(True),
        ],
    )
    bench.plot_sweep(
        title="Meta Bench",
        description="""## All Combinations of Variable Sweeps and Resulting Plots
This uses bencher to display all the combinations of plots bencher is able to produce""",
        input_vars=[
            bch.p("float_vars", [2, 3]),
            "categorical_vars",
        ],
    )

    #     bench.plot_sweep(
    #         title="Meta Bench",
    #         description="""## All Combinations of Variable Sweeps and Resulting Plots
    # This uses bencher to display all the combinations of plots bencher is able to produce""",
    #         input_vars=[
    #             bch.p("float_vars", [2, 3]),
    #             "categorical_vars",
    #         ],
    #         const_vars=[
    #             dict(level=3)
    #             # BenchMeta.param.sample_with_repeats.with_const(2),
    #             # BenchMeta.param.categorical_vars.with_const(2),
    #             # BenchMeta.param.sample_over_time.with_const(True),
    #         ],
    #     )

    return bench


if __name__ == "__main__":
    example_meta().report.show()



from __future__ import annotations
from typing import Optional
import panel as pn
from param import Parameter
import hvplot.xarray  # noqa pylint: disable=duplicate-code,unused-import
import hvplot.pandas  # noqa pylint: disable=duplicate-code,unused-import
import xarray as xr

from bencher.results.panel_result import PanelResult
from bencher.results.bench_result_base import ReduceType

from bencher.plotting.plot_filter import VarRange
from bencher.variables.results import ResultVar


class HvplotResult(PanelResult):
    def to_explorer(self) -> pn.pane.Pane:
        """Produces a hvplot explorer instance to explore the generated dataset
        see: https://hvplot.holoviz.org/getting_started/explorer.html

        Returns:
            pn.pane.Pane: A dynamic pane for exploring a dataset
        """

        if len(self.bench_cfg.input_vars) > 0:
            return self.to_xarray().hvplot.explorer()

        # For some reason hvplot doesn't like 1D datasets in xarray, so convert to pandas which it has no problem with
        # TODO look into why this is, its probably due to how I am setting up the indexing in xarray.
        return self.to_pandas().hvplot.explorer()

    def to_histogram(self, result_var: Parameter = None, **kwargs) -> Optional[pn.pane.Pane]:
        return self.filter(
            self.to_histogram_ds,
            float_range=VarRange(0, 0),
            cat_range=VarRange(0, None),
            input_range=VarRange(0, 0),
            reduce=ReduceType.NONE,
            target_dimension=2,
            result_var=result_var,
            result_types=(ResultVar),
            **kwargs,
        )

    def to_histogram_ds(self, dataset: xr.Dataset, result_var: Parameter, **kwargs):
        return dataset.hvplot(
            kind="hist",
            y=[result_var.name],
            ylabel="count",
            legend="bottom_right",
            widget_location="bottom",
            title=f"{result_var.name} vs Count",
            **kwargs,
        )



from __future__ import annotations
import panel as pn
from panel.pane import LaTeX
from typing import Optional

pn.extension("mathjax")


def latex_text(var):
    var_str = str(var)
    return r"\text{" + var_str.replace("_", " ") + r"} \\"


def input_var_to_latex(input_var):
    vals = input_var.values()
    displayed_vals = vals
    if len(vals) > 5:
        displayed_vals = [vals[i] for i in [0, 1, 0, -2, -1]]
        displayed_vals[2] = "⋮"
    v = r"\\ ".join([str(va) for va in displayed_vals])

    latex_str = r"""\begin{array}{c}"""
    latex_str += latex_text(input_var.name)
    latex_str += f"{len(vals)}" + r"\times1 \\"
    latex_str += r"\left[ \begin{array}{c}"
    latex_str += f"{v} "
    latex_str += r"\end{array} \right] \end{array}"
    return latex_str


def result_var_to_latex(bench_cfg):
    latex_str = r"\begin{array}{c}"
    sizes = [str(len(i.values())) for i in bench_cfg.all_vars]
    if len(sizes) == 1:
        sizes.insert(0, "1")
    sizes_str = r"\times".join(reversed(sizes))
    latex_str += sizes_str + r"\\ of \\"
    latex_str += r" \left[\begin{array}{cc}"

    for rv in bench_cfg.result_vars:
        latex_str += latex_text(rv.name)
    latex_str += r"\end{array} \right]"
    latex_str += r"\end{array}"
    return latex_str


def to_latex(bench_cfg) -> Optional[pn.pane.LaTeX]:
    if len(bench_cfg.all_vars) > 0:
        latex_str = r"\["
        for i, iv in enumerate(bench_cfg.all_vars):
            latex_str += input_var_to_latex(iv)
            if i != len(bench_cfg.all_vars) - 1:
                latex_str += r"\bigtimes"

        latex_str += r"\rightarrow\quad"
        latex_str += result_var_to_latex(bench_cfg)
        latex_str += r"\]"
        # underscore represents subscript, so replace underscore with large space character \;
        return LaTeX(latex_str.replace("_", r"\;"))
    return None



import nbformat as nbf
from pathlib import Path


def convert_example_to_jupyter_notebook(filename: str, output_path: str):
    # print
    source_path = Path(filename)

    nb = nbf.v4.new_notebook()
    title = source_path.stem
    function_name = f"{source_path.stem}()"
    text = f"""# {title}"""

    code = "%%capture\n"

    example_code = source_path.read_text(encoding="utf-8")
    split_code = example_code.split("""if __name__ == "__main__":""")
    code += split_code[0]

    code += f"""
bench={function_name}
"""

    code_results = """
from bokeh.io import output_notebook
output_notebook()
bench.get_result().to_auto_plots()
"""

    nb["cells"] = [
        nbf.v4.new_markdown_cell(text),
        nbf.v4.new_code_cell(code),
        nbf.v4.new_code_cell(code_results),
    ]
    output_path = Path(f"docs/reference/{output_path}/ex_{title}.ipynb")
    output_path.parent.mkdir(parents=True, exist_ok=True)
    output_path.write_text(nbf.writes(nb), encoding="utf-8")


if __name__ == "__main__":
    convert_example_to_jupyter_notebook(
        "/workspaces/bencher/bencher/example/inputs_1D/example_1_in_1_out.py", "1D"
    )
    convert_example_to_jupyter_notebook(
        "/workspaces/bencher/bencher/example/inputs_1D/example_1_in_2_out.py", "1D"
    )

    convert_example_to_jupyter_notebook(
        "/workspaces/bencher/bencher/example/inputs_1D/example_1_in_2_out_repeats.py", "1D"
    )



import bencher as bch


class DataSource:
    def __init__(self):
        self.data = [
            [0, 0, 0, 0],
            [1, 1, 1, 1],
            [1, 1, 1, 1],
            [2, 1, 1, 0],
            [2, 2, 0, 0],
            [2, 2, 1, 1],
        ]

        self.call_count = [0] * len(self.data)

    def call(self, index, repeat=None):
        if repeat is None:
            self.call_count[index] += 1
            repeat = self.call_count[index]
        return self.data[index][repeat - 1]


class Example1D(bch.ParametrizedSweep):
    index = bch.IntSweep(default=0, bounds=[0, 5], doc="Input angle", units="rad", samples=30)
    output = bch.ResultVar(units="v", doc="sin of theta")
    output2 = bch.ResultVar(units="v", doc="-sin of theta")

    def __init__(self, **params):
        super().__init__(**params)
        self.data1 = DataSource()
        self.data2 = DataSource()

    def __call__(self, **kwargs):
        self.update_params_from_kwargs(**kwargs)
        self.output = self.data1.call(self.index)
        self.output2 = -self.data2.call(self.index)
        return super().__call__(**kwargs)


def example_1_in_2_out_repeats(
    run_cfg: bch.BenchRunCfg = None, report: bch.BenchReport = None
) -> bch.Bench:
    """This example shows how to sample a 1 dimensional float variable and plot the result of passing that parameter sweep to the benchmarking function"""

    from importlib.metadata import version

    print(version("holobench"))
    if run_cfg is None:
        run_cfg = bch.BenchRunCfg()
    run_cfg.repeats = 4
    bench = Example1D().to_bench(run_cfg, report)
    bench.plot_sweep()
    return bench


if __name__ == "__main__":
    run_config = bch.BenchRunCfg()
    reprt = bch.BenchReport()
    example_1_in_2_out_repeats(run_config, reprt)
    reprt.show()
